/* interp.c    interpreter

	som-v16 add file i/o 			7 Nov 2005
	t-code som-v17					1 Jan 2006 (Happy new year!)
	add syscall stop				20 Jan 2006
	som-v19v    tx-code				20 Apr 2007
	som-v5		t2-code				7 Oct 2009
	som-v51		t2-code 64-bit      16 Dec 2010
*/

#include "compile.h"

/* run time environment
   define code, data, stack segments here

   system area  1 .. MAXSYS-1
   1..63           local
   100..300        link vectors and string
   390..700        small constants -10..300
   1000... MAXDS   global
*/

int M[MAXMEM];					// data and code segment
int SS[MAXSS];					// stack segment

int *CS = &M[MAXDS];			// code segment
int DS = MAXSYS;				// current free data segment

extern int call_load;			// ads for "call loadfile"
extern int cs_ads;				// base ads for CS[]

PRIVATE FILE *filep[MAXFILE]; 	// store file pointer
PRIVATE int fp = 1;				// current state of computation
PRIVATE int sp = 3;
// PRIVATE int param[100];
PRIVATE char strbuf[255];		// buffer for fprints

// switch CS between user and system
#define  user_cs()  CS = &M[M[cs_ads]]
#define  sys_cs()   CS = &M[MAXDS]

int noi;

void initfile(void){
	int i, a;
	filep[0] = stdin;
	filep[1] = stdout;
	filep[2] = stderr;
	for(i=3; i<MAXFILE; i++)
		filep[i] = NULL;
	// init small constants
	a = 390;
	for(i=-10; i<=300; i++,a++)
		M[a] = i;
}

// get a new file pointer, return 0 if full
PRIVATE int newfp(void){
	int i;
	for(i=3; i<MAXFILE; i++)
		if(filep[i]==NULL) return i;
	return 0;
}

// static alloc memory of size a
int xalloc(int a){
	int b;
//	printf("alloc %d DS %d\n",a,DS);
	b = DS;
	if( DS + a >= MAXDS )
		error("alloc: out of memory");
	DS += a;
	return b;
}

// gets() and put it in M[ads]
PRIVATE void xgets(FILE *fp, int ads){
	char buf[MAXLBUF], *c;
	int i;
	c = fgets(buf,MAXLBUF,fp);
	if( c != NULL ){
		for(i=0; buf[i] != 0; i++, ads++)
			M[ads] = (int)buf[i] & 255;
		M[ads] = 0;
	}else{
		M[ads] = EOF_CHAR;
		M[ads+1] = 0;
	}
}

// save fp, and 1..n register
PRIVATE void saveContext(int n){
	int i, k;
	k = sp + 1;
	for(i=1; i<=n; i++){
		SS[k] = M[i];
		k++;
	}
	SS[k] = fp;
	fp = k;
	sp = k;
}

PRIVATE void restoreContext(int n){
	int i, k;
	k = fp - 1;
	for(i=n; i>=1; i--){
		M[i] = SS[k];
		k--;
	}
	sp = k;
	fp = SS[fp];
}

// exec syscall n, options in: a, b
PRIVATE void xsys(int n, int a, int b){
	int f;
	char name[NAMELEN];

	switch(n){
	case 1: printf("%d",a); break;		// pr int
	case 2: printf("%c",a); break;		// pr char
	case 3: M[RETVAL]=getchar(); break;	// getchar
	case 4: xgets(stdin,a); break;		// gets
	case 5:								// fopen fname mode
		strunpack(name,a);				// file name
		f = newfp();
		if(f == 0) error("cannot open, too many files");
		switch(b){
		case 0: filep[f] = fopen(name,"r"); break;
		case 1: filep[f] = fopen(name,"w"); break;
		case 2: filep[f] = fopen(name,"r+"); break;
		default: error("unknown file mode");
		}
		if( filep[f] == NULL ) error("cannot open file");
		M[RETVAL] = f;					// return file pointer
		break;
	case 6:								// fclose fp
		fclose(filep[a]);
		filep[a] = NULL;
		break;
	case 7:	fprintf(filep[a],"%d",b); break; 	// fpr fp num
	case 8:	fprintf(filep[a],"%c",b); break; 	// fpr fp c
	case 9:	M[RETVAL] = fgetc(filep[a]); break; // fgetc fp
	case 10: xgets(filep[a],b); break;			// fgets fp buf
	case 11:									// fprints fp str
		strunpack(strbuf,b);
		fprintf(filep[a],"%s",strbuf);
		break;
	case 12:					// eval from som
		user_cs();				// switch to user space
		saveContext(3);			// fix 3 reg, should use b
		eval(a);
		restoreContext(3);
		sys_cs();				// switch back
		break;

//	case 13:  runflag = 0;  	// halt, this is in eval()

	case 14: M[RETVAL] = xalloc(a);	break;
	case 15:					// exec load from user space
		sys_cs();
		M[1] = a;				// filename
		eval(call_load);
		user_cs();
		break;
//	case 16: M[RETVAL] = lex(filep[a]);	break; // lex

	default:
//		printf("syscall %d ",n);
		error("undefined syscall");
	}
}

// instruction format t2-code   op c a b : c = a op b

#define nxt()		ip+=2

void eval(int ip){
	int runflag;
	int i, arity, nlocal, op;
	int a2, c, a, b, d;

	runflag = 1;
	while(runflag ) {
//		if( noi++ > 3000000 ) break;
//		noi++;

		op = CS[ip] & 0x03f;	// d:26 op:6, b:32
		d = CS[ip] >> 6;
		c = d & 0x03ff;			// a:16 c:10
		a = d >> 10;
		b = CS[ip+1];

// ---- start exec -----

	switch(op){
	case tAdd: M[c] = M[a] + M[b]; nxt(); break;
	case tSub: M[c] = M[a] - M[b]; nxt(); break;
	case tMul: M[c] = M[a] * M[b]; nxt(); break;
	case tDiv: M[c] = M[a] / M[b]; nxt(); break;
	case tMod: M[c] = M[a] % M[b]; nxt(); break;
	case tAnd: M[c] = M[a] & M[b]; nxt(); break;
	case tOr:  M[c] = M[a] | M[b]; nxt(); break;
	case tXor: M[c] = M[a] ^ M[b]; nxt(); break;
	case tShl: M[c] = M[a] << M[b]; nxt(); break;
	case tShr: M[c] = M[a] >> M[b]; nxt(); break;
	case tEq:  M[c] = M[a] == M[b]; nxt(); break;
	case tNe:  M[c] = M[a] != M[b]; nxt(); break;
	case tLt:  M[c] = M[a] <  M[b]; nxt(); break;
	case tLe:  M[c] = M[a] <= M[b]; nxt(); break;
	case tGt:  M[c] = M[a] >  M[b]; nxt(); break;
	case tGe:  M[c] = M[a] >= M[b]; nxt(); break;

	case tLdx: M[c] = M[M[a] + M[b]]; nxt(); break;
	case tStx: M[M[a] + M[b]] = M[c]; nxt(); break;

	case tJmp: ip += a; break;
	case tJf:  if(M[c] == FALSE) ip += a; else nxt(); break;
	case tJt:  if(M[c] != FALSE) ip += a; else nxt(); break;
	case tJeq: if(M[c] == M[b]) ip += a; else nxt(); break;
	case tJne: if(M[c] != M[b]) ip += a; else nxt(); break;
	case tJlt: if(M[c] <  M[b]) ip += a; else nxt(); break;
	case tJle: if(M[c] <= M[b]) ip += a; else nxt(); break;
	case tJgt: if(M[c] >  M[b]) ip += a; else nxt(); break;
	case tJge: if(M[c] >= M[b]) ip += a; else nxt(); break;

	case tCall:					// fun arity (a) fs (c)
		arity = CS[b] >> 16;
		nlocal = (CS[b] >> 6) & 0x03f;
		if (fp + nlocal + FUDGE > MAXSS)
			error("stack overflow");

//		1..arity..nlocal     p1..pn,1..arity..nlocal,ip',fp'
//		  M                 ^     ^      SS               ^
//		  reg              fp    sp                    new fp

		// first two param to stack
		a2 = sp - arity;		// at first param
		SS[a2+1] = M[a];
		SS[a2+2] = M[c];
		// create new frame
		d = sp + nlocal + 2;	// new fp
		SS[d] = fp;				// save fp
		SS[d-1] = ip + 2;		// ret ads
/*
		// save reg and pass param
		for(i = 1; i <= arity; i++){
			SS[sp+i] = M[i];
			M[i] = SS[fp+i];
		}
		// save reg the rest
		for( ; i <= nlocal; i++)
			SS[sp+i] = M[i];
*/
		for(i = 1; i <= nlocal; i++)	// save reg
			SS[sp+i] = M[i];
		for(i = 1; i <= arity; i++)		// pass param
			M[i] = SS[a2+i];
		fp = d;
		sp = fp + 2;			// spare first two slots
		ip = b + 2;
		break;
	case tRet:					// ret src (a) fs (c)
		M[RETVAL] = M[a];		// ret value
		sp = fp - c - 2;
		for(i=1; i<=c; i++)		// restore reg
			M[i] = SS[sp+i];
		ip = SS[fp-1];			// restore ip
		fp = SS[fp];			// restore fp
		break;
	case tCallt:
		arity = CS[b] >> 16;
		// first two param to stack
		a2 = sp - arity;		// at first param
		SS[a2+1] = M[a];
		SS[a2+2] = M[c];
		// pass param
		for(i = 1; i <= arity; i++)
			M[i] = SS[a2+i];
		sp = fp + 2;
		ip = b + 2;
		break;
	case tNot:   M[c] = M[a] == FALSE; nxt(); break;
	case tMov:   M[d] = M[b]; nxt(); break;
	case tPush:  sp++; SS[sp] = M[a]; nxt(); break;
	case tEfor:  M[c]++; if(M[c] <= M[b]) ip += a; else nxt(); break;
	case tCase:						// a lo, b hi
		a2 = M[c];
		if( a2 >= a && a2 <= b )
			ip += CS[ip+3+a2-a];	// jump vector
		else
			ip += CS[ip+2];			// end of jtab
		break;
	case tSys:
		if(c == 13) runflag = 0;
		else xsys(c,M[b],M[a]);
		nxt();
		break;

	default: error("undefined opcode");
	}

// ------ end exec --------

	}

//	printf("noi %d\n",noi);
}



