/* interp.c    interpreter

	som-v16 add file i/o 			7 Nov 2005
	t-code som-v17					1 Jan 2006 (Happy new year!)
	add syscall stop				20 Jan 2006
	som-v19v    tx-code				20 Apr 2007
	som-v5		t2-code				7 Oct 2009
	som-v51		t2-code 64-bit      16 Dec 2010
	integrate with vm-som vm.c
	som-v52     release				10 Nov 2012
*/

#include "compile.h"

/* run time environment
   define code, data, stack segments here

   system area  1 .. MAXSYS-1
   1..63           local
   100..300        link vectors and string
   390..700        small constants -10..300
   1000... MAXDS   global
*/

int M[MAXMEM];					// data and code segment
int SS[MAXSS];					// stack segment
int *CS = &M[MAXDS];			// code segment
int DS = MAXSYS;				// current free data segment
int fp = 1;						// current state of computation
int sp = 3;
int noi;

extern int call_load;			// ads for "call loadfile"
extern int cs_ads;				// base ads for CS[]

PRIVATE FILE *filep[MAXFILE]; 	// store file pointer
PRIVATE char strbuf[255];		// buffer for fprints

// switch CS between user and system
#define  user_cs()  CS = &M[M[cs_ads]]
#define  sys_cs()   CS = &M[MAXDS]

void initfile(void){
	int i, a;
	filep[0] = stdin;
	filep[1] = stdout;
	filep[2] = stderr;
	for(i=3; i<MAXFILE; i++)
		filep[i] = NULL;
	// init small constants
	a = 390;
	for(i=-10; i<=300; i++,a++)
		M[a] = i;
}

// get a new file pointer, return 0 if full
PRIVATE int newfp(void){
	int i;
	for(i=3; i<MAXFILE; i++)
		if(filep[i]==NULL) return i;
	return 0;
}

// gets() and put it in M[ads]
PRIVATE void xgets(FILE *fp, int ads){
	char buf[MAXLBUF], *c;
	int i;
	c = fgets(buf,MAXLBUF,fp);
	if( c != NULL ){
		for(i=0; buf[i] != 0; i++, ads++)
			M[ads] = (int)buf[i] & 255;
		M[ads] = 0;
	}else{
		M[ads] = EOF_CHAR;
		M[ads+1] = 0;
	}
}

// save fp, and 1..n register
PRIVATE void saveContext(int n){
	int i, k;
	k = sp + 1;
	for(i=1; i<=n; i++){
		SS[k] = M[i];
		k++;
	}
	SS[k] = fp;
	fp = k;
	sp = k+2;		// spare 2 slots
}

PRIVATE void restoreContext(int n){
	int i, k;
	k = fp - 1;
	for(i=n; i>=1; i--){
		M[i] = SS[k];
		k--;
	}
	sp = k;
	fp = SS[fp];
}

// exec syscall n, options in: a, b
void xsys(int n, int a, int b){
	int f, c;
	char name[NAMELEN];

	switch(n){
	case 1: printf("%d",a);	break;		// pr int
	case 2: printf("%c",a); break;		// pr char
	case 3: M[RETVAL]=getchar(); break;	// getchar
	case 4: xgets(stdin,a); break;		// gets
	case 5:								// fopen fname mode
		strunpack(name,a);				// file name
		f = newfp();
		if(f == 0) error("cannot open, too many files");
		switch(b){
		case 0: filep[f] = fopen(name,"r"); break;
		case 1: filep[f] = fopen(name,"w"); break;
		case 2: filep[f] = fopen(name,"r+"); break;
		default: error("unknown file mode");
		}
		if( filep[f] == NULL ) error("cannot open file");
		M[RETVAL] = f;					// return file pointer
		break;
	case 6:								// fclose fp
		fclose(filep[a]);
		filep[a] = NULL;
		break;
	case 7:	fprintf(filep[a],"%d",b); break; 	// fpr fp num
	case 8:	fprintf(filep[a],"%c",b); break; 	// fpr fp c
	case 9:	M[RETVAL] = fgetc(filep[a]); break; // fgetc fp
	case 10: xgets(filep[a],b); break;			// fgets fp buf
	case 11:									// fprints fp str
		strunpack(strbuf,b);
		fprintf(filep[a],"%s",strbuf);
		break;
	case 12:					// eval from som
		user_cs();				// switch to user space
		saveContext(3);			// fix 3 reg, should use b
		eval(a);
		restoreContext(3);
		sys_cs();				// switch back
		break;

//	case 13:  runflag = 0;  	// halt, this is in eval()

	case 14: M[RETVAL] = xalloc(a);	break;
	case 15:					// exec load from user space
		sys_cs();
		M[1] = a;				// filename
		eval(call_load);
		user_cs();
		break;
//	case 16: M[RETVAL] = lex(filep[a]);	break; // lex
	case 17:
		fscanf(filep[a],"%d",&c);	// fgetint fp
		M[RETVAL] = c;
		break;
	default:
		error("undefined syscall");
	}
}
